!! acoustic modeling
!! variable density
program aWFk_par
  use sep
  implicit none

  logical :: verb
  logical :: abc
  integer :: nz,nx,nt,nw, nsz, nsx
  integer :: iz,ix,it,iw, isz, isx
  real    :: dz,dx,dt,dw, dsz, dsx
  real    :: oz,ox,ot,ow, osz, osx
  real    :: energy
  integer :: bndx,bndz
  real    :: idx,idz, idt2
  real    :: c1, c2, c3, c4,  c0
  real    :: c1x,c2x,c3x,c4x
  real    :: c1z,c2z,c3z,c4z
  integer :: iop,nop
  integer :: inode, node, nodes
  integer, external  :: omp_get_num_procs 
  integer, external  :: omp_get_thread_num 

  logical :: wfld_out
  logical :: vp_out

  !! wavefield
  character(len=128)            :: wfld_
  real, dimension(:,:), allocatable :: up,um,uo,ud
  real, dimension(:,:), allocatable :: U 
  real, dimension(:,:,:), allocatable :: UU 

  !! source
  real, dimension(:),   allocatable :: sou, wvl, tmp_sou
  integer :: izsou,ixsou

  !! modulus: K = r * v^2
  character(len=128)            :: kk_
  real, dimension(:,:), allocatable :: kk, vp

  !! density
  character(len=128)            :: ro_
  real, dimension(:,:), allocatable :: ro

  !! velocity...only output for test
  character(len=128)            :: vp_


  real    :: delp
  integer :: jsnap

  !----------------------------------------------------------------
  real, dimension(:,:), allocatable :: tt
  real, dimension(:,:), allocatable :: iroxp,irozp
  real, dimension(:,:), allocatable :: iroxm,irozm
  !----------------------------------------------------------------
  real, dimension(:), allocatable :: bndz0,bndz1,bndx0,bndx1
  real, dimension(:), allocatable :: xtaper,ztaper
  real                        :: xtapco,ztapco
  !----------------------------------------------------------------  
  nop = 4

  kk_ ='kk'; ro_ ='ro'; wfld_='wfld'; vp_='vp'

  call sep_init()
  call from_param("verb",verb,.false.);   call from_param("abc",abc,.true.);  call from_param("jsnap",jsnap,15);  
  call from_param("wfld_out",wfld_out,.false.); call from_param("vp_out",vp_out,.false.)
  !!padding 
  call from_param("bndz",bndz,nop); if(bndz<nop) bndz=nop; call from_param("bndx",bndx,nop); if(bndx<nop) bndx=nop
  !!source/ shot params
  call from_history("n1",nw);  call from_param("nt",nt,nw); call from_param("dt",dt,.0001)
  !! modulus
  call from_aux(kk_,"n1",nz,1 ); call from_aux(kk_,"d1",dz,1.); call from_aux(kk_,"o1",oz,0.)
  call from_aux(kk_,"n2",nx,1 ); call from_aux(kk_,"d2",dx,1.); call from_aux(kk_,"o2",ox,0.)

  !! source location
  call from_param("nsz",nsz,1 ); call from_param("dsz",dsz,1.); call from_param("osz",osz,oz)
  call from_param("nsx",nsx,1 ); call from_param("dsx",dsx,1.); call from_param("osx",osx,ox)

  !! taper coefficients
  call from_param("xtapco",xtapco,0.0002); call from_param("ztapco",ztapco,0.0002)

  !----------------------------------------------------------------
  
  !!padding !! output files
  call to_history("n1",nx ); call to_history("d1",dx ); call to_history("o1",ox);  call to_history("n2",nt ); call to_history("d2",dt ); call to_history("o2",0.)
  call to_history("n3",nsx); call to_history("d3",dsx); call to_history("o3",osx);  call to_history("n4",nsz); call to_history("d4",dsz); call to_history("o4",osz)
if (wfld_out)then
  call auxputch("n1",'i',nz,trim(wfld_)); call auxputch("d1",'f',dz,trim(wfld_)); call auxputch("o1",'f',oz,trim(wfld_))
  call auxputch("n2",'i',nx,trim(wfld_)); call auxputch("d2",'f',dx,trim(wfld_)); call auxputch("o2",'f',ox,trim(wfld_))
  call auxputch("n3",'i',nt/jsnap,trim(wfld_)); call auxputch("d3",'f',dt*jsnap,trim(wfld_)); call auxputch("o3",'f',      0.,trim(wfld_))
  call auxputch("n4",'i',nsx,trim(wfld_)); call auxputch("d4",'f',dsx,trim(wfld_)); call auxputch("o4",'f',osx,trim(wfld_))
  call auxputch("n5",'i',nsz,trim(wfld_)); call auxputch("d5",'f',dsz,trim(wfld_)); call auxputch("o5",'f',osz,trim(wfld_))
endif
  !----------------------------------------------------------------
  call from_param("nodes" ,nodes, 1); node =omp_get_num_procs(); nodes=min(node-1,nodes) 
  call omp_set_num_threads(nodes); write(0,*) 'using', nodes, 'of' , node
  call sep_close()

  !----------------------------------------------------------------
  !----------------------------------------------------------------
  !!wld save the wavefield here
  allocate( U(nx,nt) )
  allocate( UU(nx,nt,nsx) )
!  call data_init(nx, nt, nsx, nsz)
  !----------------------------------------------------------------
  !----------------------------------------------------------------
  !----------------------------------------------------------------
  !! expand domain
  nz = nz + 2*bndz; nx = nx + 2*bndx
  allocate(up(nz,nx),um(nz,nx),uo(nz,nx),ud(nz,nx))
  allocate( ro(nz,nx))
  allocate( kk(nz,nx)); allocate( vp(nz,nx))
  allocate(sou(nt), wvl(nt) )
  !----------------------------------------------------------------  
  !! material
  call sep_read( kk(1+bndz:nz-bndz,1+bndx:nx-bndx), kk_)
  do iz=1,bndz; kk(   iz  ,:) = kk( 1+bndz,:); kk(nz-iz+1,:) = kk(nz-bndz,:); end do !modulus
  do ix=1,bndx; kk(:,   ix  ) = kk(:, 1+bndx); kk(:,nx-ix+1) = kk(:,nx-bndx); end do

  call sep_read( ro(1+bndz:nz-bndz,1+bndx:nx-bndx), ro_)
  do iz=1,bndz; ro(   iz  ,:) = ro( 1+bndz,:); ro(nz-iz+1,:) = ro(nz-bndz,:); end do
  do ix=1,bndx; ro(:,   ix  ) = ro(:, 1+bndx); ro(:,nx-ix+1) = ro(:,nx-bndx); end do

  vp = sqrt(kk/ro)

  if (vp_out)then
        call auxputch("n1",'i',nz,trim(vp_)); call auxputch("d1",'f',dz,trim(vp_)); call auxputch("o1",'f',oz,trim(vp_))
        call auxputch("n2",'i',nx,trim(vp_)); call auxputch("d2",'f',dx,trim(vp_)); call auxputch("o2",'f',ox,trim(vp_))
        call sep_write(vp, vp_, esize=4)
  endif
  write(0,*) 'max------------------' 
  write(0,*) 'max_vel', maxval(vp),  'min_vel', minval(vp); 
  
  idx = 1./dx; idz = 1./dz; idt2= 1./dt**2
  !----------------------------------------------------------------
  !! source derivative (displacement to acceleration)
  sou=0; wvl=0.
  write(0,*)size(sou),nw,nt
  call sep_read(sou(1:nw))

  do iw=1+1,nt-1
     wvl(iw) = (-2.*sou(iw) + sou(iw+1) + sou(iw-1))
  end do
  wvl = wvl * idt2
!  wvl = sou
  !----------------------------------------------------------------
  !! 1/ro
  allocate(iroxp(nz,nx),iroxm(nz,nx)); allocate(irozp(nz,nx),irozm(nz,nx)); allocate(  tt(nz,nx))
  do iz=1+1,nz-1
     irozp(iz,:) = 2./(ro(iz+1,:)+ro(iz  ,:))
     irozm(iz,:) = 2./(ro(iz  ,:)+ro(iz-1,:))
  end do
  irozp( 1,:) = irozp(   2,:)
  irozp(nz,:) = irozp(nz-1,:)
  irozm( 1,:) = irozm(   2,:)
  irozm(nz,:) = irozm(nz-1,:)

  do ix=1+1,nx-1
     iroxp(:,ix) = 2./(ro(:,ix+1)+ro(:,ix  ))
     iroxm(:,ix) = 2./(ro(:,ix  )+ro(:,ix-1))
  end do
  iroxp(:, 1) = iroxp(:,   2)
  iroxp(:,nx) = iroxp(:,nx-1)
  iroxm(:, 1) = iroxm(:,   2)
  iroxm(:,nx) = iroxm(:,nx-1)
  !----------------------------------------------------------------
  if(abc) then
     !! one-way absorbing BC setup
     allocate(bndz0(nx),bndz1(nx))
     do ix=1,nx
        delp = vp( 1+nop,ix) *dt/dz
        bndz0(ix) = (1.-delp)/(1.+delp)

        delp = vp(nz-nop,ix) *dt/dz
        bndz1(ix) = (1.-delp)/(1.+delp)
     end do

     allocate(bndx0(nz),bndx1(nz))
     do iz=1,nz
        delp = vp(iz, 1+nop) *dt/dx
        bndx0(iz) = (1.-delp)/(1.+delp)

        delp = vp(iz,nx-nop) *dt/dx
        bndx1(iz) = (1.-delp)/(1.+delp)
     end do
  end if
  !----------------------------------------------------------------
  if(abc) then
     !! taper setup
     allocate(ztaper(bndz))
     do iz=1,bndz
        ztaper(iz) = exp(-ztapco * iz*iz)
     end do

     allocate(xtaper(bndx))
     do ix=1,bndx
        xtaper(ix) = exp(-xtapco * ix*ix)
        write(0,*)xtaper(ix)
     end do
  end if
  !----------------------------------------------------------------
  !! derivative coefficients
  c1= 1.      *1225./1024.
  c2=-1./15.  *1225./1024.
  c3= 1./125. *1225./1024.
  c4=-1./1715.*1225./1024.

  c1x = c1 * idx; c1z = c1 * idz
  c2x = c2 * idx; c2z = c2 * idz
  c3x = c3 * idx; c3z = c3 * idz
  c4x = c4 * idx; c4z = c4 * idz

  !----------------------------------------------------------------
  do isz=1,nsz;    izsou = bndz + 1+((osz+(isz-1)*dsz) - oz)/dz
  if (allocated(uo))deallocate(uo); 
  if (allocated(um))deallocate(um);  
  if (allocated(ud))deallocate(ud); 
  if (allocated(up))deallocate(up); 

  if (allocated(U))deallocate(U);   
  if (allocated(tt))deallocate(tt);   
!$OMP parallel do ordered private(inode, isx, ixsou, it, ix, iz , tt, uo, ud, um, up, U, iop) schedule(dynamic,1)
     do isx=1,nsx; ixsou = bndx + 1+((osx+(isx-1)*dsx) - ox)/dx
        inode=omp_get_thread_num();  
          if (allocated(uo))deallocate(uo); 
          if (allocated(um))deallocate(um);  
          if (allocated(ud))deallocate(ud); 
          if (allocated(up))deallocate(up)
          if (allocated(U))deallocate(U);   
          if (allocated(tt))deallocate(tt);   
          allocate(up(nz,nx),um(nz,nx),uo(nz,nx),ud(nz,nx),tt(nz,nx),U(nx-2*bndx,nt))
        write(0,*) "sou",(isz-1)*nsx+isx,"/",nsx*nsz,"@", osz+(isz-1)*dsz, osx+(isx-1)*dsx, '@', inode   

        !tmp_sou = wvl / ro(izsou,ixsou)
        !! MAIN loop
        up=0;um=0;uo=0;ud=0;
        do it=1,nt
           !! 4th order derivatives (GOOD)
           do iz=nop,nz-nop ! forward derivative
              tt(iz,:) = &
              c1z * (uo(iz+1,:) - uo(iz-0,:)) + &
              c2z * (uo(iz+2,:) - uo(iz-1,:)) + &
              c3z * (uo(iz+3,:) - uo(iz-2,:)) + &
              c4z * (uo(iz+4,:) - uo(iz-3,:))
           end do
           tt = tt * iroxp
           do iz=1+nop,nz-nop ! backward derivative
              ud(iz,:) = &
              c1z * (tt(iz+0,:) - tt(iz-1,:)) + &
              c2z * (tt(iz+1,:) - tt(iz-2,:)) + &
              c3z * (tt(iz+2,:) - tt(iz-3,:)) + &
              c4z * (tt(iz+3,:) - tt(iz-4,:))
           end do

           do ix=nop,nx-nop ! forward derivative
              tt(:,ix) = &
              c1x * (uo(:,ix+1) - uo(:,ix-0)) + &
              c2x * (uo(:,ix+2) - uo(:,ix-1)) + &
              c3x * (uo(:,ix+3) - uo(:,ix-2)) + &
              c4x * (uo(:,ix+4) - uo(:,ix-3))
           end do
           tt = tt * irozp
           do ix=1+nop,nx-nop ! backward derivative
              ud(:,ix) = &
              ud(:,ix) + &
              c1x * (tt(:,ix+0) - tt(:,ix-1)) + &
              c2x * (tt(:,ix+1) - tt(:,ix-2)) + &
              c3x * (tt(:,ix+2) - tt(:,ix-3)) + &
              c4x * (tt(:,ix+3) - tt(:,ix-4))
           end do

           ud = ud * kk

           !! acceleration source
!           ud(izsou,ixsou) = &
!           ud(izsou,ixsou) - sou(it)
           ud(izsou,ixsou) = &
           ud(izsou,ixsou) - wvl(it)
           if(abc) then
              !! taper
              do iz=1,nz
                 uo(iz,        1:bndx) = uo(iz,        1:bndx) * xtaper(bndx:1:-1) 
                 uo(iz,nx-bndx+1:  nx) = uo(iz,nx-bndx+1:  nx) * xtaper(1:bndx:+1)
                 um(iz,        1:bndx) = um(iz,        1:bndx) * xtaper(bndx:1:-1) 
                 um(iz,nx-bndx+1:  nx) = um(iz,nx-bndx+1:  nx) * xtaper(1:bndx:+1)
                 ud(iz,        1:bndx) = ud(iz,        1:bndx) * xtaper(bndx:1:-1) 
                 ud(iz,nx-bndx+1:  nx) = ud(iz,nx-bndx+1:  nx) * xtaper(1:bndx:+1)
              end do
              do ix=1,nx
                 uo(        1:bndz,ix) = uo(        1:bndz,ix) * ztaper(bndz:1:-1) 
                 uo(nz-bndz+1:  nz,ix) = uo(nz-bndz+1:  nz,ix) * ztaper(1:bndz:+1)
                 um(        1:bndz,ix) = um(        1:bndz,ix) * ztaper(bndz:1:-1) 
                 um(nz-bndz+1:  nz,ix) = um(nz-bndz+1:  nz,ix) * ztaper(1:bndz:+1)
                 ud(        1:bndz,ix) = ud(        1:bndz,ix) * ztaper(bndz:1:-1) 
                 ud(nz-bndz+1:  nz,ix) = ud(nz-bndz+1:  nz,ix) * ztaper(1:bndz:+1)
              end do
           end if

           !! time step
           up = 2*uo - um + ud * dt**2
           um =   uo
           uo =   up
           if(abc) then
              !! one-way absorbing BC
              do iop=1,nop
                 uo( 1+nop-iop  ,:) = &
                 um( 1+nop-iop+1,:) + bndz0 * (um( 1+nop-iop,:)-uo( 1+nop-iop+1,:))
                 uo(nz-nop+iop  ,:) = &
                 um(nz-nop+iop-1,:) + bndz1 * (um(nz-nop+iop,:)-uo(nz-nop+iop-1,:))
                 uo(:, 1+nop-iop  ) = &
                 um(:, 1+nop-iop+1) + bndx0 * (um(:, 1+nop-iop)-uo(:, 1+nop-iop+1))
                 uo(:,nx-nop+iop  ) = &
                 um(:,nx-nop+iop-1) + bndx1 * (um(:,nx-nop+iop)-uo(:,nx-nop+iop-1))
              end do
           end if
           
           if(mod(it,jsnap)==0 .and. wfld_out) then
              call sep_write(uo(1+bndz:nz-bndz,1+bndx:nx-bndx),wfld_)
           end if
          ! call sep_write(uo(1+bndz,1+bndx:nx-bndx)) 
            U(:,it) = uo(1+bndz,1+bndx:nx-bndx) 
 !         if(verb .and. inode==0) write(0,*) it,sum(uo)
        end do
        !UU(:,:,isx)=U
        !$OMP ORDERED
                write(0,*)'writing....shot',isx,sum(U)
                call sep_write(U) 
        !UU(:,:,isx)=U
        !call data_io(U,nx-2*bndx,nt,nsx,isx)
        !$OMP END ORDERED 
     end do
!$OMP end parallel do
  end do
  !call sep_write(UU) 
  call exit(0)
end program AWFk_par


!===================================================
  subroutine data_init(nx, nt, nsx, nsz)
  use sep
    complex, allocatable, dimension(:) :: ctmp
    integer                            :: i, nloop
    character(len=256)                  ::  data_

    write(0,*)'Initializing data...'
    data_='out'
        allocate(ctmp(nsz*nx*nt))
        write(0,*), nx,nt,nsx,nsz
        ctmp = 0
        nloop = nsx
        do i=1, nloop
                write(0,*)i, ' of ', nloop
                call sep_write(ctmp, data_, esize=4)
        end do
    deallocate(ctmp)
  end subroutine
!##################################################################
subroutine data_io(data, nx, nt, nsx, isx)
    real, dimension(:,:), pointer       :: data
    integer                             :: nx, nt, isx, nsx 
    integer, dimension(3)               :: n_cube, j_wind, n_wind, f_wind
    integer                             :: n_dim_cube, err
    character :: y
    character(len=256)                  ::  data_
    data_="outtag"

    n_dim_cube = 3
    n_cube = (/nx , nt, nsx   /)
    j_wind = (/1  , 1 , 1     /)
    n_wind = (/nx , nt, 1     /)
    f_wind = (/0  , 0 , isx-1 /)
    write(0,*) n_cube
    write(0,*) n_wind
    write(0,*) f_wind
    write(0,*) j_wind

    write(0,*)'writing shot:',isx,' of ', nsx, '...'
       err = srite_window("out", n_dim_cube, n_cube, n_wind, f_wind, j_wind, 4, data)
       if (err/=0) call seperr("problem writing data")
       write(0,*)'done.'
  end subroutine

